import Buffer from './ring-buffer.js'

export const Button = Object.freeze({
    South: 0,
    East: 1,
    West: 2,
    North: 3,
    BumperLeft: 4,
    BumperRight: 5,
    TriggerLeft: 6,
    TriggerRight: 7,
    Select: 8,
    Start: 9,
    AnalogLeft: 10,
    AnalogRight: 11,
    DPadUp: 12,
    DPadDown: 13,
    DPadLeft: 14,
    DPadRight: 15,
    Logo: 16
})

export const Axis = Object.freeze({
    LeftX: 0,
    LeftY: 1,
    RightX: 2,
    RightY: 3
})

export const InputState = Object.freeze({
    Disabled: 0,
    Waiting: 1,
    Started: 2,
    Performed: 3,
    Canceled: 4
})

export const Device = Object.freeze({
    Any: 0,
    KeyboardAndMouse: 1,
    Gamepad: 2
})

class InputEvent extends CustomEvent {
    constructor(value, type, source, data) {
        super('oninput', {
            detail: {
                value,
                type,
                source,
                ...data
            }
        })
    }
}

class GamepadEvent extends InputEvent {
    constructor(value, source) {
        super(value, Device.Gamepad, source)
    }
}

class KeyboardEvent extends InputEvent {
    constructor(value, source, state, modifiers) {
        super(value, Device.KeyboardAndMouse, source, {
            modifiers,
            state
        })
    }
}

class InputDevice extends EventTarget { }

class Gamepads extends InputDevice {
    #indices
    #raf

    constructor() {
        super()

        this.#indices = new Set()
        this.#raf = undefined

        window.addEventListener('gamepadconnected', e => {
            this.#indices.add(e.gamepad.index)
        })

        window.addEventListener('gamepaddisconnected', e => {
            this.#indices.delete(e.gamepad.index)
        })
    }

    static get Button() {
        return Button
    }

    static get Axis() {
        return Axis
    }

    startPolling() {
        this.#poll()
    }

    stopPolling() {
        cancelAnimationFrame(this.#raf)
    }

    #poll() {
        this.#raf = requestAnimationFrame(() => {
            this.#testInputs()
            this.#poll()
        })
    }

    #testInputs() {
        const gamepads = navigator.getGamepads()
        for (const index of this.#indices) {
            const gamepad = gamepads[index]
            for (const button of gamepad.buttons) {
                this.dispatchEvent(new GamepadEvent(button, gamepad))
            }
        }
    }
}

export const KeyState = Object.freeze({
    Up: 0,
    Down: 1
})

export class KeyModifier {
    static Alt = 1
    static Ctrl = 2
    static Meta = 4
    static Shift = 8

    constructor() {
        throw new TypeError("can't use 'new' with an enum")
    }

    static fromString(value) {
        if (Array.isArray(value)) {
            return value.reduce((acc, x) => acc | KeyModifier.fromString(x), 0)
        }

        switch (value.toLowerCase()) {
            case "alt": return 1
            case "ctrl": return 2
            case "meta": return 4
            case "shift": return 8
        }
    }

    static fromEvent(e) {
        let result = 0

        if (e.altKey) { result |= KeyModifier.Alt }
        if (e.ctrlKey) { result |= KeyModifier.Ctrl }
        if (e.metaKey) { result |= KeyModifier.Meta }
        if (e.shiftKey) { result |= KeyModifier.Shift }

        return result
    }

    static has(flags, value) {
        return flags & value
    }

    static set(flags, value) {
        return flags | value
    }

    static toString(value) {
        let result = []

        if (value & 1) { result += "Alt" }
        if (value & 2) { result += "Ctrl" }
        if (value & 4) { result += "Meta" }
        if (value & 8) { result += "Shift" }

        return result.join(',')
    }
}

class Keyboard extends InputDevice {
    constructor() {
        super()

        document.addEventListener('keydown', e => {
            const state = e.repeat ? InputState.Performed : InputState.Started

            this.dispatchEvent(
                new KeyboardEvent(e.code, e, state, KeyModifier.fromEvent(e))
            )
        })

        document.addEventListener('keyup', e => {
            this.dispatchEvent(
                new KeyboardEvent(e.code, e, InputState.Canceled, KeyModifier.fromEvent(e))
            )
        })
    }
}

const filterProp = (prop, value, xs) => xs.filter(x => x[prop] === value)

class BaseAction {
    #values

    constructor(values) {
        this.#values = new Set(values)
    }

    get values() {
        return this.#values
    }

    get device() {
        return Device.Any
    }

    *find(inputs) {
        for (const value of this.#values) {
            for (const input of inputs) {
                if (input.value === value) {
                    yield input
                }
            }
        }
    }

    findFirst(inputs) {
        return this.find(inputs).next()
    }

    test(input) {
        return this.findFirst(input) ? true : false
    }
}

export class KeyboardAction extends BaseAction {
    #modifiers

    constructor(values) {
        const parts = values.split('+')
        super(parts.slice(-1))
        this.#modifiers = KeyModifier.fromString(parts.slice(0, -1))
    }

    get device() {
        return Device.KeyboardAndMouse
    }

    test(input) {
        const filtered = input.filter(v => v.type === this.device && v.state === InputState.Started)
        for (const match of this.find(filtered)) {
            if (match.modifiers == this.#modifiers) {
                return true
            }
        }
        return false
    }
}

export class GamepadAction extends BaseAction {
    constructor(values) {
        super(values)
    }

    get device() {
        return Device.Gamepad
    }

    test(input) {
        super.test(
            filterProp('device', this.device, input)
        )
    }
}

export class Action {
    static keyboard(input) {
        return new KeyboardAction(input)
    }

    static gamepad(input) {
        return new GamepadAction(input)
    }
}


export class ActionMap {
    constructor(devices, bufferSize = 16) {
        this.buffer = Buffer(bufferSize)
        this.actions = new Map()
        this.devices = !devices ? this.initializeDefaults() : devices

        this.devices.forEach(device => device.addEventListener('oninput', this.#handle.bind(this)))
    }

    initializeDefaults() {
        return [
            new Keyboard(),
            new Gamepads()
        ]
    }

    bind(action, callback) {
        this.actions.set(action, callback)
        return this
    }

    unbind(action) {
        this.actions.delete(action)
    }

    #handle(e) {
        this.buffer.enqueue(e.detail)
        this.#test()
    }

    #test() {
        for (const [action, callback] of this.actions.entries()) {
            if (action.test(this.buffer.valueOf())) {
                this.buffer.clear()
                callback()
            }
        }
    }

    clearBuffer() {
        this.buffer.clear()
    }
}

